"""Plots the script flow."""
from typing import Dict, List, Union

import matplotlib.pyplot as plt
import networkx as nx
import PIL
from typeguard import typechecked


def visualise_script_flow(
    G: nx.DiGraph, app_name: str, app_version: str
) -> None:
    """Visualises the script flow."""

    G.nodes[0]["pos"] = [0, 0]
    set_cyclical_node_coords(evaluated_nodes=[], G=G, start_nodename=0, y=0)
    set_node_images(G=G, app_name=app_name, app_version=app_version)
    plot_coordinated_graph(
        G=G,
    )


def set_cyclical_node_coords(
    evaluated_nodes: List[int],
    G: nx.DiGraph,
    start_nodename: int,
    y: int,
) -> nx.DiGraph:
    """Sets the coordinates for a cyclical graph."""
    neighbors = list(G.neighbors(start_nodename))
    if "pos" not in G.nodes[start_nodename].keys():
        raise KeyError("Error, pos value of start node was not set.")
    for x, neighbor in enumerate(neighbors):
        if neighbor not in evaluated_nodes:
            if "pos" not in G.nodes[neighbor].keys():
                G.nodes[neighbor]["pos"] = [y + 1, x]
    for x, neighbor in enumerate(neighbors):
        if neighbor not in evaluated_nodes:
            evaluated_nodes.append(neighbor)
            set_cyclical_node_coords(
                evaluated_nodes=evaluated_nodes,
                G=G,
                start_nodename=neighbor,
                y=y + 1,
            )


def set_node_images(G: nx.DiGraph, app_name: str, app_version: str) -> None:
    """Sets the phone screens as node images to visualise the script flow
    between screens."""
    # Image URLs for graph nodes
    icons: Dict[str, str] = {}
    for screen_nr in G.nodes:
        icons[screen_nr] = (
            f"src/appcommander/{app_name}/V{app_version}/verified/"
            + f"{screen_nr}.png"
        )
        # TODO: verify file exists, use dummy file otherwise.
    # Load images
    images = {k: PIL.Image.open(fname) for k, fname in icons.items()}

    # Generate the computer network graph
    for nodename in G.nodes:
        G.nodes[nodename]["image"] = images[nodename]


@typechecked
def plot_coordinated_graph(
    G: Union[nx.Graph, nx.DiGraph],
) -> None:
    """Plots the networkx graph with images instead of nodes.

    :param G: The original graph on which the MDSA algorithm is ran.
    """
    pos: Dict[int, List[int]] = {}
    for nodename in G.nodes:
        pos[nodename] = G.nodes[nodename]["pos"]
    print(f"pos={pos}")

    fig = plt.figure(figsize=(1, 1))
    ax = plt.subplot(111)
    ax.set_aspect("equal")
    nx.draw_networkx_edges(G, pos, ax=ax)

    # plt.xlim(-1,10)
    # plt.ylim(-1.5,1.5)

    trans = ax.transData.transform
    trans2 = fig.transFigure.inverted().transform

    piesize = 0.3  # this is the image size
    p2 = piesize / 2.0
    for n in G:
        xx, yy = trans(pos[n])  # figure coordinates
        xa, ya = trans2((xx, yy))  # axes coordinates
        print(f"n={n},xa,ya={xa,ya}")
        a = plt.axes([xa - p2, ya - p2, piesize, piesize])
        a.set_aspect("equal")
        a.imshow(G.nodes[n]["image"])
        a.axis("off")

    ax.axis("off")
    plt.show()



"""Functions to assist a script file for an arbitrary app."""
import importlib
from typing import TYPE_CHECKING, List

import networkx as nx
from typeguard import typechecked

if TYPE_CHECKING:
    from src.appcommander.org_torproject_android.V16_6_3_RC_1.Script import (
        Script,
    )
    from src.appcommander.Screen import Screen
else:
    Script = object
    Screen = object

screen_path: str = "src.appcommander.org_torproject_android.V16_6_3_RC_1."
moduleNames = []
screen_func_names = []
modules = []
for screen_index in range(0, 8):
    module_name = f"{screen_path}screen_{screen_index}"
    moduleNames.append(module_name)
    screen_func_names.append(f"screen_{screen_index}")
    my_module = importlib.import_module(module_name)
    modules.append(my_module)


@typechecked
def create_screens(script_graph: nx.DiGraph) -> List[Screen]:
    """Creates the screens as networkx nodes."""
    screens: List[Screen] = []

    # Create the Screen objects programmatically.
    for i, module in enumerate(modules):
        # Create the function (reference) programmatically.
        # module represents the file that contains the screen function.
        # screen_func_name[i] is the name of the screen_i function in that
        # module.
        # screen_function is the actual Pythonic function handle. it is
        # like writing: screen_3 if i==3.
        screen_function = getattr(module, screen_func_names[i])
        # execute the screen function, which returns a Screen object.
        screens.append(screen_function())

    # Add the screen objects to the script graph.
    for screen in screens:
        script_graph.nodes[screen.screen_nr]["Screen"] = screen
    return screens



"""Script to run through Orbot app configuration.

Android names this app: org.torproject.android
"""

from typing import TYPE_CHECKING, Dict, List

import networkx as nx
from typeguard import typechecked

from src.appcommander.create_screens import create_screens
from src.appcommander.org_torproject_android.V16_6_3_RC_1.screen_flow import (
    Script_flow,
)

if TYPE_CHECKING:
    from src.appcommander.Screen import Screen
else:
    Screen = object


# pylint: disable=R0902
class Script:
    """Experiment manager.

    First prepares the environment for running the experiment, and then
    calls a private method that executes the experiment consisting of 4
    stages.
    """

    # pylint: disable=R0903
    # pylint: disable=R0913
    @typechecked
    def __init__(
        self,
        app_name: str,
        overwrite: bool,
        package_name: str,
        torifying_apps: Dict[str, str],
        version: str,
    ) -> None:
        self.app_name: str = app_name
        self.overwrite: bool = overwrite
        self.package_name: str = package_name
        self.torifying_apps: Dict[str, str] = torifying_apps
        self.version: str = version

        # Create placeholder for past screens.
        self.past_screens: List[int] = []

        # Generate the script screen flow as a graph and generate the screens.
        self.script_graph = Script_flow().G
        self.screens: List[Screen] = create_screens(self.script_graph)

        # Specify the start and end nodes in the graph.
        self.set_start_nodes(self.script_graph)

    @typechecked
    def set_start_nodes(self, script_graph: nx.DiGraph) -> None:
        """Sets the is_start attributes to True in the nodes that are start
        screens."""

        for nodename in script_graph.nodes:
            screen: Screen = script_graph.nodes[nodename]["Screen"]
            if screen.screen_nr in list(range(0, 8)):
                script_graph.nodes[nodename]["is_start"] = True
            else:
                script_graph.nodes[nodename]["is_start"] = False



"""Packaging logic for appcommander."""
from __future__ import annotations

import os
import sys

import setuptools

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "src"))

install_requires = [
    "",
]
setuptools.setup()
